@@ -1,5 +1,5 @@
 /*******************************************************************************
-    Copyright (c) 2015-2020 NVIDIA Corporation
+    Copyright (c) 2015-2023 NVIDIA Corporation
 
     Permission is hereby granted, free of charge, to any person obtaining a copy
     of this software and associated documentation files (the "Software"), to
@@ -57,8 +57,11 @@
 #include "uvm_tracker.h"
 #include "uvm_va_block_types.h"
 #include "uvm_linux.h"
-#include "uvmtypes.h"
+#include "uvm_types.h"
 #include "nv_uvm_types.h"
+#if UVM_IS_CONFIG_HMM()
+#include <linux/memremap.h>
+#endif
 
 typedef enum
 {
@@ -94,9 +97,18 @@ typedef enum
 {
     // Memory type for backing user pages. On Pascal+ it can be evicted.
     UVM_PMM_GPU_MEMORY_TYPE_USER,
+    // When the Confidential Computing feature is enabled, the protected flavor
+    // allocates memory out of the VPR region. When it's disabled, all flavors
+    // have no effects and are equivalent to the base type.
+    UVM_PMM_GPU_MEMORY_TYPE_USER_PROTECTED = UVM_PMM_GPU_MEMORY_TYPE_USER,
+    UVM_PMM_GPU_MEMORY_TYPE_USER_UNPROTECTED,
 
     // Memory type for internal UVM allocations. It cannot be evicted.
     UVM_PMM_GPU_MEMORY_TYPE_KERNEL,
+    // See user types for the behavior description when the Confidential
+    // Computing feature is ON or OFF.
+    UVM_PMM_GPU_MEMORY_TYPE_KERNEL_PROTECTED = UVM_PMM_GPU_MEMORY_TYPE_KERNEL,
+    UVM_PMM_GPU_MEMORY_TYPE_KERNEL_UNPROTECTED,
 
     // Number of types - MUST BE LAST.
     UVM_PMM_GPU_MEMORY_TYPE_COUNT
@@ -104,6 +116,16 @@ typedef enum
 
 const char *uvm_pmm_gpu_memory_type_string(uvm_pmm_gpu_memory_type_t type);
 
+// Returns true if the given memory type is used to back user pages.
+bool uvm_pmm_gpu_memory_type_is_user(uvm_pmm_gpu_memory_type_t type);
+
+// Returns true if the given memory type is used to back internal UVM
+// allocations.
+static bool uvm_pmm_gpu_memory_type_is_kernel(uvm_pmm_gpu_memory_type_t type)
+{
+    return !uvm_pmm_gpu_memory_type_is_user(type);
+}
+
 typedef enum
 {
     // Chunk belongs to PMA. Code outside PMM should not have access to
@@ -185,7 +207,6 @@ typedef uvm_chunk_size_t uvm_chunk_sizes_mask_t;
 
 typedef struct uvm_pmm_gpu_chunk_suballoc_struct uvm_pmm_gpu_chunk_suballoc_t;
 
-typedef struct uvm_gpu_chunk_struct uvm_gpu_chunk_t;
 struct uvm_gpu_chunk_struct
 {
     // Physical address of GPU chunk. This may be removed to save memory
@@ -198,15 +219,15 @@ struct uvm_gpu_chunk_struct
         // We use +1 in the order_base_2 calls appropriately to avoid compiler
         // warnings due to the bitfields being too narrow for the values of
         // their types.
-        uvm_pmm_gpu_memory_type_t               type : order_base_2(UVM_PMM_GPU_MEMORY_TYPE_COUNT + 1);
+        uvm_pmm_gpu_memory_type_t type : order_base_2(UVM_PMM_GPU_MEMORY_TYPE_COUNT + 1);
 
         // The eviction flag is internal and used only for root chunks. It's
         // set by the eviction path once a chunk is chosen for eviction in
         // chunk_start_eviction(). Also see the (root_)chunk_is_in_eviction()
         // helpers.
-        bool                             in_eviction : 1;
+        bool in_eviction : 1;
 
-        bool                      inject_split_error : 1;
+        bool inject_split_error : 1;
 
         // This flag is initalized when allocating a new root chunk from PMA.
         // It is set to true, if PMA already scrubbed the chunk. The flag is
@@ -214,19 +235,23 @@ struct uvm_gpu_chunk_struct
         // the caller is not required to clear it before freeing the chunk. The
         // VA block chunk population code can query it to skip zeroing the
         // chunk.
-        bool                                 is_zero : 1;
+        bool is_zero : 1;
 
-        uvm_pmm_gpu_chunk_state_t              state : order_base_2(UVM_PMM_GPU_CHUNK_STATE_COUNT + 1);
+        // This flag indicates an allocated chunk is referenced by a device
+        // private struct page PTE and therefore expects a page_free() callback.
+        bool is_referenced : 1;
 
-        size_t                             log2_size : order_base_2(UVM_CHUNK_SIZE_MASK_SIZE);
+        uvm_pmm_gpu_chunk_state_t state : order_base_2(UVM_PMM_GPU_CHUNK_STATE_COUNT + 1);
+
+        size_t log2_size : order_base_2(UVM_CHUNK_SIZE_MASK_SIZE);
 
         // Start page index within va_block
-        uvm_page_index_t         va_block_page_index : order_base_2(PAGES_PER_UVM_VA_BLOCK + 1);
+        uvm_page_index_t va_block_page_index : order_base_2(PAGES_PER_UVM_VA_BLOCK + 1);
 
         // This allows determining what PMM owns the chunk. Users of this field
         // must only use it if the owning GPU is retained.
         // TODO: Bug 2008200: Enforce single PMM instance per GPU
-        NvU32                           gpu_global_index : order_base_2(UVM_GLOBAL_ID_MAX_PROCESSORS);
+        NvU32 gpu_index : order_base_2(UVM_ID_MAX_PROCESSORS);
     };
 
     // List entry.
@@ -265,11 +290,9 @@ typedef struct uvm_gpu_root_chunk_struct
     // uvm_pmm_gpu_t::root_chunks.indirect_peer[id].dma_addrs.
     //
     // Protected by the corresponding root chunk bit lock.
-
     //
     // We can use a regular processor id because indirect peers are not allowed
     // between partitioned GPUs when SMC is enabled.
-
     uvm_processor_mask_t indirect_peers_mapped;
 } uvm_gpu_root_chunk_t;
 
@@ -301,11 +324,8 @@ typedef struct
     atomic64_t map_count;
 } uvm_gpu_root_chunk_indirect_peer_t;
 
-typedef struct
+typedef struct uvm_pmm_gpu_struct
 {
-    // TODO: Bug 2008200: Remove this field and use container_of
-    uvm_gpu_t *gpu;
-
     // Sizes of the MMU
     uvm_chunk_sizes_mask_t chunk_sizes[UVM_PMM_GPU_MEMORY_TYPE_COUNT];
 
@@ -343,6 +363,12 @@ typedef struct
         // List of root chunks used by VA blocks
         struct list_head va_block_used;
 
+        // List of chunks needing to be lazily freed and a queue for processing
+        // the list. TODO: Bug 3881835: revisit whether to use nv_kthread_q_t
+        // or workqueue.
+        struct list_head va_block_lazy_free;
+        nv_kthread_q_item_t va_block_lazy_free_q_item;
+
         uvm_gpu_root_chunk_indirect_peer_t indirect_peer[UVM_ID_MAX_GPUS];
     } root_chunks;
 
@@ -365,11 +391,16 @@ typedef struct
     // The mask of the initialized chunk sizes
     DECLARE_BITMAP(chunk_split_cache_initialized, UVM_PMM_CHUNK_SPLIT_CACHE_SIZES);
 
+    bool initialized;
+
     bool pma_address_cache_initialized;
 } uvm_pmm_gpu_t;
 
+// Return containing GPU
+uvm_gpu_t *uvm_pmm_to_gpu(uvm_pmm_gpu_t *pmm);
+
 // Initialize PMM on GPU
-NV_STATUS uvm_pmm_gpu_init(uvm_gpu_t *gpu, uvm_pmm_gpu_t *pmm);
+NV_STATUS uvm_pmm_gpu_init(uvm_pmm_gpu_t *pmm);
 
 // Deinitialize the PMM on GPU
 void uvm_pmm_gpu_deinit(uvm_pmm_gpu_t *pmm);
@@ -400,18 +431,23 @@ struct page *uvm_gpu_chunk_to_page(uvm_pmm_gpu_t *pmm, uvm_gpu_chunk_t *chunk);
 // Allocates num_chunks chunks of size chunk_size in caller-supplied array (chunks).
 //
 // Returned chunks are in the TEMP_PINNED state, requiring a call to either
-// uvm_pmm_gpu_unpin_temp or uvm_pmm_gpu_free. If a tracker is passed in, all
+// uvm_pmm_gpu_unpin_allocated, uvm_pmm_gpu_unpin_referenced, or
+// uvm_pmm_gpu_free. If a tracker is passed in, all
 // the pending operations on the allocated chunks will be added to it
 // guaranteeing that all the entries come from the same GPU as the PMM.
 // Otherwise, when tracker is NULL, all the pending operations will be
 // synchronized before returning to the caller.
 //
 // Each of the allocated chunks list nodes (uvm_gpu_chunk_t::list) can be used
-// by the caller until the chunk is unpinned (uvm_pmm_gpu_unpin_temp) or freed
-// (uvm_pmm_gpu_free). If used, the list node has to be returned to a valid
-// state before calling either of the APIs.
+// by the caller until the chunk is unpinned (uvm_pmm_gpu_unpin_allocated,
+// uvm_pmm_gpu_unpin_referenced) or freed (uvm_pmm_gpu_free). If used, the list
+// node has to be returned to a valid state before calling either of the APIs.
 //
 // In case of an error, the chunks array is guaranteed to be cleared.
+//
+// If the memory returned by the PMM allocator cannot be physically addressed,
+// the MMU interface provides user chunk mapping and unmapping functions
+// (uvm_mmu_chunk_map/unmap) that enable virtual addressing.
 NV_STATUS uvm_pmm_gpu_alloc(uvm_pmm_gpu_t *pmm,
                             size_t num_chunks,
                             uvm_chunk_size_t chunk_size,
@@ -424,16 +460,26 @@ NV_STATUS uvm_pmm_gpu_alloc(uvm_pmm_gpu_t *pmm,
 //
 // Internally calls uvm_pmm_gpu_alloc() and sets the state of all chunks to
 // allocated on success.
-NV_STATUS uvm_pmm_gpu_alloc_kernel(uvm_pmm_gpu_t *pmm,
-                                   size_t num_chunks,
-                                   uvm_chunk_size_t chunk_size,
-                                   uvm_pmm_alloc_flags_t flags,
-                                   uvm_gpu_chunk_t **chunks,
-                                   uvm_tracker_t *out_tracker);
+//
+// If Confidential Computing is enabled, this helper allocates protected kernel
+// memory.
+static NV_STATUS uvm_pmm_gpu_alloc_kernel(uvm_pmm_gpu_t *pmm,
+                                          size_t num_chunks,
+                                          uvm_chunk_size_t chunk_size,
+                                          uvm_pmm_alloc_flags_t flags,
+                                          uvm_gpu_chunk_t **chunks,
+                                          uvm_tracker_t *out_tracker)
+{
+    return uvm_pmm_gpu_alloc(pmm, num_chunks, chunk_size, UVM_PMM_GPU_MEMORY_TYPE_KERNEL, flags, chunks, out_tracker);
+}
 
 // Helper for allocating user memory
 //
-// Simple wrapper that just uses UVM_PMM_GPU_MEMORY_TYPE_USER for the memory type.
+// Simple wrapper that just uses UVM_PMM_GPU_MEMORY_TYPE_USER for the memory
+// type.
+//
+// If Confidential Computing is enabled, this helper allocates protected user
+// memory.
 static NV_STATUS uvm_pmm_gpu_alloc_user(uvm_pmm_gpu_t *pmm,
                                         size_t num_chunks,
                                         uvm_chunk_size_t chunk_size,
@@ -444,10 +490,17 @@ static NV_STATUS uvm_pmm_gpu_alloc_user(uvm_pmm_gpu_t *pmm,
     return uvm_pmm_gpu_alloc(pmm, num_chunks, chunk_size, UVM_PMM_GPU_MEMORY_TYPE_USER, flags, chunks, out_tracker);
 }
 
-// Unpin a temporarily pinned chunk and set its reverse map to a VA block
+// Unpin a temporarily pinned chunk, set its reverse map to a VA block, and
+// mark it as allocated.
 //
 // Can only be used on user memory.
-void uvm_pmm_gpu_unpin_temp(uvm_pmm_gpu_t *pmm, uvm_gpu_chunk_t *chunk, uvm_va_block_t *va_block);
+void uvm_pmm_gpu_unpin_allocated(uvm_pmm_gpu_t *pmm, uvm_gpu_chunk_t *chunk, uvm_va_block_t *va_block);
+
+// Unpin a temporarily pinned chunk, set its reverse map to a VA block, and
+// mark it as referenced.
+//
+// Can only be used on user memory.
+void uvm_pmm_gpu_unpin_referenced(uvm_pmm_gpu_t *pmm, uvm_gpu_chunk_t *chunk, uvm_va_block_t *va_block);
 
 // Frees the chunk. This also unpins the chunk if it is temporarily pinned.
 //
